"""Example OctoAI service scaffold: Wav2Vec."""
import torch
from transformers import Wav2Vec2ForCTC, Wav2Vec2Processor

from octoai.service import Service
from octoai.types import Audio, Text


class Wav2VecService(Service):
    """An OctoAI service extends octoai.service.Service."""

    def setup(self):
        """Download model weights to disk."""
        self.processor = Wav2Vec2Processor.from_pretrained(
            "facebook/wav2vec2-base-960h"
        )
        self.model = Wav2Vec2ForCTC.from_pretrained("facebook/wav2vec2-base-960h")

    def infer(self, audio: Audio) -> Text:
        """Perform inference with the model."""
        audio_array, sampling_rate = audio.to_numpy()
        input_values = self.processor(
            audio_array,
            sampling_rate=sampling_rate,
            return_tensors="pt",
            padding="longest",
        ).input_values
        logits = self.model(input_values).logits
        predicted_ids = torch.argmax(logits, dim=-1)
        transcription = self.processor.batch_decode(predicted_ids)

        return Text(text=transcription[0])



"""Example OctoAI service scaffold: Flan-T5-Small."""
from transformers import T5ForConditionalGeneration, T5Tokenizer

from octoai.service import Service
from octoai.types import Text


class T5Service(Service):
    """An OctoAI service extends octoai.service.Service."""

    def setup(self):
        """Download model weights to disk."""
        self.tokenizer = T5Tokenizer.from_pretrained("google/flan-t5-small")
        self.model = T5ForConditionalGeneration.from_pretrained("google/flan-t5-small")

    def infer(self, prompt: Text) -> Text:
        """Perform inference with the model."""
        inputs = self.tokenizer(prompt.text, return_tensors="pt")
        outputs = self.model.generate(**inputs)
        response = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)

        return Text(text=response[0])



"""Example OctoAI service scaffold: YOLOv8."""
import json
from typing import List

import numpy as np
from PIL import Image as PImage
from pydantic import BaseModel, Field
from ultralytics import YOLO

from octoai.service import Service
from octoai.types import Image

# This model has an elaborate prediction schema
# (labels, probabilities, and box coordinates).
# Since we can't rely on `octoai` predefined types
# we define our own to capture that structure.


class Box(BaseModel):
    """Represents corners of a detection box."""

    x1: float
    x2: float
    y1: float
    y2: float


class Detection(BaseModel):
    """Represents a detection."""

    name: str
    class_: int = Field(..., alias="class")
    confidence: float
    box: Box


class YOLOResponse(BaseModel):
    """Response includes list of detections and rendered image."""

    detections: List[Detection]
    image_with_boxes: Image


class YOLOv8Service(Service):
    """An OctoAI service extends octoai.service.Service."""

    def setup(self):
        """Download model weights to disk."""
        self.model = YOLO("yolov8l.pt")

    def infer(self, image: Image) -> YOLOResponse:
        """Perform inference with the model."""
        image_pil = image.to_pil()
        output = self.model(image_pil)

        # no Python dict available directly from the model
        detections = json.loads(output[0].tojson())

        # get the rendered image with bounding boxes and labels
        # - Result.plot() generates BGR; flip to RGB
        img_out_numpy = np.flip(output[0].plot(), axis=2)
        img_out_pil = PImage.fromarray(img_out_numpy)

        # Return detection data and a rendered image with boxes
        return YOLOResponse(
            detections=[Detection(**d) for d in detections],
            image_with_boxes=Image.from_pil(img_out_pil),
        )



"""
Server for OctoAI endpoints created with the ``octoai`` CLI.

Developers that want to create an endpoint should not use
this module directly. Instead, they should use the ``octoai``
command-line interface, which directs them to implement the
``octoai.service.Service`` class and use the ``octoai`` CLI to help
build and deploy their endpoint.
"""
import importlib
import inspect
import json
import time
from enum import Enum
from typing import Any, Dict, Optional

import click
import uvicorn
from fastapi import FastAPI
from fastapi.responses import JSONResponse

from .service import (
    ResponseAnalytics,
    Service,
    inspect_input_types,
    inspect_output_types,
)

_OCTOAI_SERVICE_MODULE = "octoai.service"
_OCTOAI_BASE_SERVICE_CLASS = "Service"


class HealthStatus(Enum):
    """HTTP status codes for health check."""

    STARTING = 503
    UNKNOWN = 500
    READY = 200


class Server:
    """
    Server for OctoAI endpoints created with the ``octoai`` CLI.

    Developers that want to create an endpoint should not use
    this class directly. Instead, they should use the ``octoai``
    command-line interface, which directs them to implement the
    ``octoai.service.Service`` class and use the ``octoai`` CLI to
    help build and deploy their endpoint.
    """

    def __init__(self, service: Service):
        self.app = FastAPI()
        self.service = service
        self.status = HealthStatus.STARTING

        Input = inspect_input_types(service)
        Output = inspect_output_types(service)

        @self.app.get("/healthcheck")
        def health() -> JSONResponse:
            return JSONResponse(
                status_code=self.status.value, content={"status": self.status.name}
            )

        @self.app.get("/")
        def root() -> JSONResponse:
            return JSONResponse(
                status_code=200,
                content={
                    "docs": "/docs",
                    "openapi": "/openapi.json",
                },
            )

        @self.app.post(
            "/infer",
            response_model=Output,
        )
        def infer(request: Input):
            infer_args = {k: v for k, v in request}

            # track time elapsed in nanoseconds only while app is not asleep
            start_process = time.process_time_ns()
            # track time elapsed in nanoseconds including any sleep time
            start_perf = time.perf_counter_ns()
            prediction = service.infer(**infer_args)
            inference_time_ms = (time.process_time_ns() - start_process) / 1e6
            performance_time_ms = (time.perf_counter_ns() - start_perf) / 1e6

            return Output(
                output=prediction,
                analytics=ResponseAnalytics(
                    inference_time_ms=inference_time_ms,
                    performance_time_ms=performance_time_ms,
                ),
            )

    def get_api_schema(self) -> Dict[str, Any]:
        """Return the Open API schema for the underlying service."""
        return self.app.openapi()

    def init(self):
        """Initialize the underlying service."""
        try:
            self.service.setup()
            self.status = HealthStatus.READY
        except Exception as e:
            self.status = HealthStatus.UNKNOWN
            raise e

    def run(self, port: int):
        """Run the server exposing the underlying service."""
        uvicorn.run(self.app, host="0.0.0.0", port=port)


def load_service(module_name: str, class_name: Optional[str]) -> Service:
    """Load the service implementation."""
    try:
        module = importlib.import_module(module_name)

        if class_name is not None:
            # if service class is provided, instantiate it
            class_ = getattr(module, class_name)
        else:
            # if service class not provided, look for it
            class_ = None
            for name, class_obj in inspect.getmembers(module, inspect.isclass):
                for class_base in class_obj.__bases__:
                    if (
                        class_base.__module__ == _OCTOAI_SERVICE_MODULE
                        and class_base.__name__ == _OCTOAI_BASE_SERVICE_CLASS
                    ):
                        class_ = class_obj
                        break

            if class_ is None:
                raise ValueError(
                    f"Module '{module_name}' contains no classes extending "
                    f"base '{_OCTOAI_SERVICE_MODULE}.{_OCTOAI_BASE_SERVICE_CLASS}'"
                )

        click.echo(f"Using service in {module_name}.{class_.__name__}.")

        return class_()
    except ModuleNotFoundError:
        error_msg = f"Module '{module_name}' not found. "
        if module_name == "service":
            error_msg += "Ensure your service is defined in service.py."
        raise ValueError(error_msg)


@click.group(name="server")
@click.option("--service-module", default="service")
@click.option("--service-class", default=None)
@click.pass_context
def server(ctx, service_module, service_class):
    """CLI for OctoAI server."""
    click.echo("octoai server")
    ctx.ensure_object(dict)
    service = load_service(service_module, service_class)
    ctx.obj["server"] = Server(service)


@server.command()
@click.option("--output-file", default=None)
@click.pass_context
def api_schema(ctx, output_file):
    """Generate OpenAPI schema for the given service."""
    click.echo("api-schema")
    server = ctx.obj["server"]
    schema = server.get_api_schema()

    if output_file:
        with open(output_file, "w") as f:
            json.dump(schema, f, indent=2)
    else:
        click.echo(json.dumps(schema, indent=2))


@server.command()
@click.pass_context
def setup_service(ctx):
    """Run the setup code for the given service."""
    click.echo("setup-service")
    server = ctx.obj["server"]
    server.init()


@server.command()
@click.pass_context
@click.option("--port", type=int, default=8080)
def run(ctx, port):
    """Run the server for the given service."""
    click.echo("run")
    server = ctx.obj["server"]
    server.init()
    server.run(port)


if __name__ == "__main__":
    server(obj={})



"""
Type definitions to help communicate with endpoints.

These type definitions help with routine format conversions
that are necessary when transferring binary files (such as
images or audio) over HTTP. These type definitions can be useful
both when creating endpoints (implementing ``octoai.service.Service``
as directed by the ``octoai`` command-line interface) and when
communicating with live endpoints using the client SDK.
"""

import base64
import os
from io import BytesIO
from typing import Any, Dict, Tuple

import httpx
import soundfile
from numpy.typing import ArrayLike
from PIL import Image as PImage
from pydantic import BaseModel, Field


def file_b64encode(file_bytes: bytes) -> str:
    """Encode binary file as a base64 string.

    :param file_bytes: contents of a binary file
    :type file_bytes: bytes
    :return: contents of a binary file encoded as a base64 string
    :rtype: str
    """
    return base64.b64encode(file_bytes).decode("ascii")


def file_b64decode(file_b64: str) -> bytes:
    """Decode a base64 string into a binary file.

    :param file_b64: contents of a binary file encoded as a base64 string
    :type file_b64: str
    :return: contents of a binary file
    :rtype: bytes
    """
    return base64.b64decode(bytes(file_b64, "ascii"))


class Text(BaseModel):
    """Text type for models that accept or return text."""

    text: str


class Image(BaseModel):
    """Image helpers for models that accept or return images.

    To create an instance of this type, users can select the appropriate
    ``Image.from_...()`` method depending on the format of their input.

    :param BaseModel: base model
    :type BaseModel: :class:`BaseModel`
    :raises ValueError: ``from_file`` method failed to load image.
    :raises ValueError: ``from_url`` method was unable to reach the url.
    :return: base64 encoded image
    :rtype: str
    """

    image_b64: str = Field(description="base64-encoded image file")

    @classmethod
    def from_base64(cls, b64: str):
        """Create from base64 encoded string, such as that returned from an HTTP call.

        See also ``Image.from_endpoint_response()``.

        :param b64: contents of a binary image file as a base64 encoded string
        :type b64: str
        :return: Image object
        :rtype: :class:`Image`
        """
        return cls(image_b64=b64)

    @classmethod
    def from_endpoint_response(cls, resp_dict: Dict[str, Any], key: str):
        """Create from an endpoint response, such as an endpoint that produces images.

        :param resp_dict: a response from an OctoAI endpoint that produces images
        :type resp_dict: Dict[str, Any]
        :param key: the key name in the response that contains an encoded image
        :return: Image object
        :rtype: :class:`Image`
        """
        if key in resp_dict:
            return cls(image_b64=resp_dict[key]["image_b64"])
        elif "output" in resp_dict and key in resp_dict["output"]:
            return cls(image_b64=resp_dict["output"][key]["image_b64"])

        raise ValueError(f"{key} not in resp_dict")

    @classmethod
    def from_pil(cls, image_pil: PImage, format="JPEG"):
        """Create from Pillow image object.

        A file format is required since the Pillow image object is
        serialized to a binary image file. The default is "JPEG".

        :param image_pil: image in PIL format
        :type image_pil: PIL.Image
        :param format: target file format, defaults to "JPEG"
        :type format: str, optional
        :return: Image object
        :rtype: :class:`Image`
        """
        buffer = BytesIO()
        image_pil.save(buffer, format=format)
        return cls(image_b64=file_b64encode(buffer.getvalue()))

    @classmethod
    def from_file(cls, image_file: str):
        """Create from local file.

        :param image_file: path to local image file
        :type image_file: str
        :raises ValueError: image_file not found at provided path
        :return: Image object
        :rtype: :class:`Image`
        """
        if not os.path.isfile(image_file):
            raise ValueError(f"File {image_file} does not exist")

        with open(image_file, "rb") as fd:
            return cls(image_b64=file_b64encode(fd.read()))

    @classmethod
    def from_url(cls, image_url: str, follow_redirects=False):
        """Create from image URL.

        :param image_url: url leading to an image
        :type image_url: str
        :param follow_redirects: whether to follow URL redirects
        :type follow_redirects: bool
        :raises ValueError: there was an error reaching the target url
        :return: Image object
        :rtype: :class:`Image`
        """
        resp = httpx.get(image_url, follow_redirects=follow_redirects)
        if resp.status_code != 200:
            raise ValueError(f"status {resp.status_code} ({image_url})")

        return cls(image_b64=file_b64encode(resp.content))

    def is_valid(self):
        """Check if this is a valid image.

        :return: True if valid, False if invalid
        :rtype: bool
        """
        try:
            self.to_pil().verify()
            return True
        except Exception:
            return False

    def to_pil(self) -> PImage:
        """Convert to PIL Image.

        :return: Pillow image object
        :rtype: PIL.Image
        """
        return PImage.open(BytesIO(file_b64decode(self.image_b64)))

    def to_file(self, file_name: str):
        """Save image to disk.

        :param file_name: file path
        :type file_name: str
        """
        with open(file_name, "wb") as fd:
            fd.write(file_b64decode(self.image_b64))


class Audio(BaseModel):
    """Audio helpers for models that accept or return audio.

    To create an instance of this type, users can select the appropriate
    ``Audio.from_...()`` method depending on the format of their input.

    :param BaseModel: base model
    :type BaseModel: :class:`BaseModel`
    :raises ValueError: ``from_file`` method failed to load file
    :raises ValueError: ``from_url`` method was unable to reach the url.
    :return: base64 encoded audio
    :rtype: str
    """

    audio_b64: str = Field(description="base64-encoded audio file")

    @classmethod
    def from_base64(cls, b64: str):
        """Create from base64 encoded string, such as that returned from an HTTP call.

        See also ``Audio.from_endpoint_response()``.

        :param b64: contents of a binary audio file as a base64 encoded string
        :type b64: str
        :return: Audio object
        :rtype: :class:`Audio`
        """
        return cls(audio_b64=b64)

    @classmethod
    def from_endpoint_response(cls, resp_dict: Dict[str, Any], key: str):
        """Create from an endpoint response, such as an endpoint that produces audio.

        :param resp_dict: a response from an OctoAI endpoint that produces audio
        :type resp_dict: Dict[str, Any]
        :param key: the key name in the response that contains an encoded audio file
        :return: Audio object
        :rtype: :class:`Audio`
        """
        if key in resp_dict:
            return cls(audio_b64=resp_dict[key]["audio_b64"])
        elif "output" in resp_dict and key in resp_dict["output"]:
            return cls(audio_b64=resp_dict["output"][key]["audio_b64"])

        raise ValueError(f"{key} not in resp_dict")

    @classmethod
    def from_numpy(cls, data: ArrayLike, sample_rate: int, format="WAV"):
        """Create from a numpy array.

        The first dimension of the array should represent audio frames (samples),
        while the second dimension should represent audio channels.

        A file format and a sample rate are needed since the audio data is
        serialized to a binary audio file. The default format is "WAV", and the
        sample rate is required.

        :param data: numpy array with audio data (frames x channels)
        :type data: ArrayLike
        :param sample_rate: samples per second taken to create signal
        :type sample_rate: int
        :param format: target format, defaults to "WAV"
        :type format: str, optional
        :return: Audio object
        :rtype: :class:`Audio`
        """
        buffer = BytesIO()
        soundfile.write(buffer, data=data, samplerate=sample_rate, format=format)
        return cls(audio_b64=file_b64encode(buffer.getvalue()))

    @classmethod
    def from_file(cls, audio_file: str):
        """Create from local file.

        :param audio_file: path to local audio file
        :type audio_file: str
        :raises ValueError: audio_file not found at provided path
        :return: Audio object
        :rtype: :class:`Audio`
        """
        if not os.path.isfile(audio_file):
            raise ValueError(f"File {audio_file} does not exist")

        with open(audio_file, "rb") as fd:
            return cls(audio_b64=file_b64encode(fd.read()))

    @classmethod
    def from_url(cls, audio_url: str, follow_redirects=False):
        """Create from audio URL.

        :param audio_url: url leading to an audio file
        :type audio_url: str
        :param follow_redirects: whether to follow URL redirects
        :type follow_redirects: bool
        :raises ValueError: there was an error reaching the target url
        :return: Audio object
        :rtype: :class:`Audio`
        """
        resp = httpx.get(audio_url, follow_redirects=follow_redirects)
        if resp.status_code != 200:
            raise ValueError(f"status {resp.status_code} ({audio_url})")

        return cls(audio_b64=file_b64encode(resp.content))

    def is_valid(self):
        """Check if this is a valid audio.

        :return: True if it's valid, false if not.
        :rtype: bool
        """
        try:
            self.to_numpy()
            return True
        except Exception:
            return False

    def to_numpy(self) -> Tuple[ArrayLike, int]:
        """Convert to numpy array.

        :return: numpy array representation (frames x channels)
        :rtype: Tuple[ArrayLike, int]
        """
        fd = BytesIO(file_b64decode(self.audio_b64))
        data, sample_rate = soundfile.read(fd)
        return (data, sample_rate)

    def to_file(self, file_name: str):
        """Save to disk.

        :param file_name: file descriptor or path.
        :type file_name: str
        """
        with open(file_name, "wb") as fd:
            fd.write(file_b64decode(self.audio_b64))



"""
Contains the base interface that OctoAI endpoints should implement.

Developers that want to create an endpoint should implement the
``Service`` class in this module as directed by the ``octoai`` command-line
interface.
"""
import inspect
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Type

from pydantic import BaseModel, Field, create_model


class Service(ABC):
    """
    The base interface that OctoAI endpoints should implement.

    Developers that want to create an endpoint should implement this
    class as directed by the ``octoai`` command-line interface.
    """

    def setup(self) -> None:
        """
        Perform service initialization.

        A common operation to include here is loading weights and making
        those available to the ``infer()`` method in a member variable.
        """
        pass

    @abstractmethod
    def infer(self, **kwargs: Any) -> Any:
        """Perform inference."""
        pass


class ResponseAnalytics(BaseModel):
    """Additional analytics metadata."""

    inference_time_ms: float = Field(
        description="Inference execution time (without pauses)"
    )
    performance_time_ms: float = Field(
        description="Inference execution time (including pauses)"
    )


def inspect_input_types(service: Service) -> Type[BaseModel]:
    """Create Pydantic input model from ``infer()`` signature."""
    args = OrderedDict()
    signature = inspect.signature(service.infer)

    if len(signature.parameters) < 1:
        raise ValueError("infer() requires at least one argument")

    for arg_name, arg in signature.parameters.items():
        if arg.annotation == inspect._empty:
            raise ValueError("infer() requires type annotations for args")

        default = None if arg.default == inspect._empty else arg.default
        args[arg_name] = (arg.annotation, default)

    return create_model(
        "Input",
        __config__=None,
        __base__=BaseModel,
        __module__=__name__,
        __validators__=None,
        **args,
    )


def inspect_output_types(service: Service) -> Type[BaseModel]:
    """Create Pydantic output model from ``infer()`` signature."""
    signature = inspect.signature(service.infer)

    if signature.return_annotation == inspect._empty:
        raise ValueError("infer() requires a return type annotation")

    rets = OrderedDict()
    rets["output"] = (signature.return_annotation, None)
    rets["analytics"] = (ResponseAnalytics, None)

    return create_model(
        "Output",
        __config__=None,
        __base__=BaseModel,
        __module__=__name__,
        __validators__=None,
        **rets,
    )



"""Client used to infer from endpoints."""

import logging
import os
from pathlib import Path
from typing import Any, Dict, Mapping, Optional

import httpx
import yaml

import octoai
from octoai import utils
from octoai.errors import OctoAIClientError, OctoAIServerError


class Client:
    """A class that allows inferences from existing endpoints.

    :param token: api token, defaults to None
    :type token: str, optional
    :param public_endpoints_url: str, url to fetch available public models,
        defaults to "https://api.octoai.cloud/v1/public-endpoints"
    :type public_endpoints_url: str
    :param config_path: path to '/.octoai/config.yaml'.  Installed in ~,
        defaults to None and will check home path
    :type config_path: Optional[str]

    Sets various headers. Gets auth token from environment if none is provided.
    """

    def __init__(
        self,
        token: Optional[str] = None,
        public_endpoints_url: Optional[
            str
        ] = "https://api.octoai.cloud/v1/public-endpoints",
        config_path: Optional[str] = None,
    ) -> None:
        """Initialize the :class: `octoai.Client` with an auth token.

        :raises OctoAIServerError: server-side failures (unreachable, etc)
        :raises OctoAIClientError: client-side failures (throttling, unset token)
        """
        self._public_endpoints: Dict[str, str] = {}
        self._public_endpoints_url = public_endpoints_url

        token = token if token else os.environ.get("OCTOAI_TOKEN", None)

        if not token:
            # Default path is ~/.octoai/config.yaml for token, can be overridden
            path = Path(config_path) if config_path else Path.home()
            try:
                with open(
                    (path / Path(".octoai/config.yaml")), encoding="utf-8"
                ) as octoai_config_yaml:
                    config_dict = yaml.safe_load(octoai_config_yaml)
                token = config_dict.get("token")
            except FileNotFoundError:
                token = None

        if not token:
            logging.warning(
                "OCTOAI_TOKEN environment variable is not set. "
                + "You won't be able to reach OctoAI endpoints."
            )

        version = octoai.__version__  # type: ignore
        headers = {
            "Content-Type": "application/json",
            "user-agent": f"octoai-{version}",
        }

        if token:
            headers["Authorization"] = f"Bearer {token}"

        self._headers = headers
        # Set all timeouts to 900 seconds to account for cold starts, latency.
        timeout = httpx.Timeout(timeout=900.0)
        self._httpx_client = httpx.Client(timeout=timeout, headers=headers)

    def _initialize_public_endpoints(self) -> None:
        """Initialize self._public_endpoints with dict of names to quickstart urls."""
        response = utils.retry(
            lambda: self._httpx_client.get(url=self._public_endpoints_url)
        )
        if response.status_code == 200:
            response_json = response.json()
            for model in response_json:
                self._public_endpoints[model["name"]] = model["endpoint"]
        else:
            self._error(response.status_code, response.text)

    def infer(self, endpoint_url: str, inputs: Mapping[str, Any]) -> Mapping[str, Any]:
        """Send a request to the given endpoint URL with inputs as request body.

        :param endpoint_url: target endpoint
        :type endpoint_url: str
        :param inputs: inputs for target endpoint
        :type inputs: Mapping[str, Any]

        :raises OctoAIServerError: server-side failures (unreachable, etc)
        :raises OctoAIClientError: client-side failures (throttling, unset token)

        :return: outputs from endpoint
        :rtype: Mapping[str, Any]
        """
        resp = utils.retry(
            lambda: self._httpx_client.post(url=endpoint_url, json=inputs)
        )
        if resp.status_code != 200:
            self._error(resp.status_code, resp.text)
        return resp.json()

    def health_check(self, endpoint_url: str, timeout: float = 900.0) -> int:
        """Check health of an endpoint using a get request.  Try until timeout.

        :param endpoint_url: URL as a str starting with https permitting get requests.
        :type endpoint_url: str
        :param timeout: Seconds before request times out, defaults to 900.
        :type timeout: float
        :return: status code from get request.  200 means ready.
        :rtype: int
        """
        resp = utils.health_check(
            lambda: self._httpx_client.get(url=endpoint_url), timeout=timeout
        )
        if resp.status_code != 200:
            self._error(resp.status_code, resp.text)
        return resp.status_code

    def _error(self, status_code: int, text: str):
        """Raise error of correct type for status code including message.

        :param status_code: HTTP status_code
        :type status_code: int
        :param text: error message from API server
        :type text: str

        :raises OctoAIServerError: server-side failures (unreachable, etc)
        :raises OctoAIClientError: client-side failures (throttling, unset token)
        """
        if status_code >= 500:
            raise OctoAIServerError(f"Server error: {status_code} {text}")
        elif status_code == 429:
            raise OctoAIClientError(f"Throttling error: {status_code} {text}")
        else:
            raise OctoAIClientError(f"Error: {status_code} {text}")

    @property
    def public_endpoints(self) -> Dict[str, str]:
        """Return dict of public endpoint names as strs to endpoint urls as strs.

        :return: Dict of public endpoint name to URL.
        :rtype: Dict[str, str]
        """
        if not self._public_endpoints:
            self._initialize_public_endpoints()
        return self._public_endpoints

    @property
    def mpt_7b(self) -> str:
        """Return the request-limited quickstart template URL for MPT-7B.

        Allows for instruction tuned text generation.

        :return: quickstart template URL for MPT-7B
        :rtype: str
        """
        return self.public_endpoints.get("mpt-7b-demo") + "/generate"

    @property
    def vicuna_7b(self) -> str:
        """Return the request-limited quickstart template URL for Vicuna-7B.

        Allows for instruction tuned text generation.

        :return: quickstart template URL for Vicuna-7B
        :rtype: str
        """
        return self.public_endpoints.get("vicuna-7b-demo") + "/generate"

    @property
    def whisper(self) -> str:
        """Return the request-limited quickstart template URL for Whisper.

        Allows speech recognition.

        :return: quickstart template URL for Whisper
        :rtype: str
        """
        return self.public_endpoints.get("whisper-demo") + "/predict"

    @property
    def stable_diffusion(self) -> str:
        """Return the request-limited quickstart template URL for Stable Diffusion.

        Allows for generating an image when given a prompt.

        :return: quickstart template URL for Stable Diffusion
        :rtype: str
        """
        return self.public_endpoints.get("stable-diffusion-demo") + "/predict"



"""Example OctoAI service scaffold: Hello World."""
from octoai.service import Service
from octoai.types import Text


class HelloService(Service):
    """An OctoAI service extends octoai.service.Service."""

    def setup(self):
        """Perform intialization."""
        print("Setting up.")

    def infer(self, prompt: Text) -> Text:
        """Perform inference."""
        return prompt
